import functools

import pytest

from glotter.settings import Settings
from glotter.source import get_sources


def project_test(project_type):
    def decorator(func):
        Settings().add_test_mapping(project_type, func)

        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapper

    return decorator


def project_fixture(project_type):
    sources = get_sources(Settings().source_root).get(project_type)
    return pytest.fixture(
        scope="module",
        params=sources,
        ids=[
            f"{source.language}/{source.name}{source.extension}" for source in sources
        ],
    )



import re
import sys

import pytest

from glotter.source import get_sources, filter_sources
from glotter.settings import Settings


def test(args):
    test_args = ["-n", "auto"] if args.parallel else []
    if not (args.language or args.project or args.source):
        _run_pytest_and_exit(*test_args)

    all_tests = _collect_tests()
    sources_by_type = filter_sources(args, get_sources(Settings().source_root))
    for project_type, sources in sources_by_type.items():
        for source in sources:
            test_args += _get_tests(project_type, all_tests, source)

    if not test_args:
        _error_and_exit("No tests were found")

    _run_pytest_and_exit(*test_args)


def _get_tests(project_type, all_tests, src=None):
    test_functions = Settings().get_test_mapping_name(project_type)
    tests = []
    for test_func in test_functions:
        if src is not None:
            test_id = f"{src.language}/{src.name}{src.extension}"
            pattern = rf"^(\w/?)*\.py::{test_func}\[{re.escape(test_id)}(-.*)?\]$"
        else:
            pattern = rf"^(\w/?)*\.py::{test_func}\[.+\]$"
        tests.extend(
            [tst for tst in all_tests if re.fullmatch(pattern, tst) is not None]
        )
    return tests


def _run_pytest_and_exit(*args):
    args = ["-v"] + list(args)
    code = pytest.main(args=args)
    sys.exit(code)


def _error_and_exit(msg):
    print(msg)
    sys.exit(1)


class TestCollectionPlugin:
    def __init__(self):
        self.collected = []

    def pytest_collection_modifyitems(self, items):
        for item in items:
            self.collected.append(item.nodeid)


def _collect_tests():
    print(
        "============================= collect test totals =============================="
    )
    plugin = TestCollectionPlugin()
    pytest.main(["-qq", "--collect-only"], plugins=[plugin])
    return plugin.collected



from concurrent.futures import ThreadPoolExecutor

from glotter.source import get_sources, filter_sources
from glotter.settings import Settings
from glotter.containerfactory import ContainerFactory


def download(args):
    sources_by_type = filter_sources(args, get_sources(Settings().source_root))
    containers = {
        source.test_info.container_info.image: source.test_info.container_info
        for sources in sources_by_type.values()
        for source in sources
    }

    if args.parallel:
        with ThreadPoolExecutor(max_workers=4) as executor:
            executor.map(
                lambda source: _download_image_from_source(source, True),
                containers.values(),
            )
    else:
        for container_info in containers:
            _download_image_from_source(container_info)


def _download_image_from_source(container_info, parallel=False):
    ContainerFactory().get_image(container_info, parallel=parallel)



import os
from warnings import warn

import yaml

from glotter.project import Project, AcronymScheme
from glotter.containerfactory import Singleton


class Settings(metaclass=Singleton):
    def __init__(self):
        self._project_root = os.getcwd()
        self._parser = SettingsParser(self._project_root)
        self._projects = self._parser.projects
        self._source_root = self._parser.source_root or self._project_root
        self._test_mappings = {}

    @property
    def projects(self):
        return self._projects

    @property
    def project_root(self):
        return self._project_root

    @property
    def source_root(self):
        return self._source_root

    @source_root.setter
    def source_root(self, value):
        self._source_root = value or self._project_root

    @property
    def test_mappings(self):
        return self._test_mappings

    def get_test_mapping_name(self, project_type):
        mappings = self._test_mappings.get(project_type)
        if mappings:
            return [func.__name__ for func in mappings]
        return []

    def add_test_mapping(self, project_type, func):
        if project_type not in self._projects:
            raise KeyError(f"Project type {project_type} was not found in glotter.yml")

        if project_type not in self._test_mappings:
            self._test_mappings[project_type] = []
        self._test_mappings[project_type].append(func)

    def verify_project_type(self, name):
        try:
            return name.lower() in self.projects
        except KeyError as e:
            raise Exception(  # pylint: disable=broad-exception-raised
                f'glotter.yml does not contain project name "{name}"'
            ) from e


class SettingsParser:
    def __init__(self, project_root):
        self._project_root = project_root
        self._yml_path = None
        self._yml = None
        self._acronym_scheme = None
        self._projects = None
        self._source_root = None
        self._yml_path = self._locate_yml()
        if self._yml_path is not None:
            self._yml = self._parse_yml()
            self.parse_settings_section()
            self.parse_projects_section()
        else:
            warn(f'.glotter.yml not found in directory "{project_root}"')

    def parse_settings_section(self):
        if self._yml is not None:
            self._acronym_scheme = self._parse_acronym_scheme()
            self._source_root = self._parse_source_root()

    def parse_projects_section(self):
        if self.yml is not None:
            self._projects = self._parse_projects()

    @property
    def project_root(self):
        return self._project_root

    @property
    def yml_path(self):
        return self._yml_path

    @property
    def yml(self):
        return self._yml

    @property
    def source_root(self):
        return self._source_root

    @property
    def acronym_scheme(self):
        return self._acronym_scheme

    @property
    def projects(self):
        return self._projects

    def _parse_acronym_scheme(self):
        if "settings" not in self._yml or "acronym_scheme" not in self._yml["settings"]:
            return None

        scheme = self._yml["settings"]["acronym_scheme"].lower()
        return AcronymScheme[scheme]

    def _parse_source_root(self):
        return self._parse_root("source_root")

    def _parse_root(self, key):
        if "settings" not in self._yml or key not in self._yml["settings"]:
            return None

        path = self._yml["settings"][key]
        if os.path.isabs(path):
            return path

        yml_dir = os.path.dirname(self._yml_path)
        return os.path.abspath(os.path.join(yml_dir, path))

    def _parse_projects(self):
        projects = {}
        if "projects" in self._yml:
            for k, v in self._yml["projects"].items():
                project_name = k.lower()
                project = Project(
                    words=v.get("words"),
                    requires_parameters=v.get("requires_parameters"),
                    acronyms=v.get("acronyms"),
                    acronym_scheme=v.get("acronym_scheme") or self._acronym_scheme,
                )
                projects[project_name] = project

        return projects

    def _parse_yml(self):
        with open(self._yml_path, "r", encoding="utf-8") as f:
            contents = f.read()

        return yaml.safe_load(contents)

    def _locate_yml(self):
        for root, _, files in os.walk(self._project_root):
            if ".glotter.yml" in files:
                path = os.path.abspath(root)
                return os.path.join(path, ".glotter.yml")

        return None



from enum import Enum, auto


class NamingScheme(Enum):
    hyphen = auto()
    underscore = auto()
    camel = auto()
    pascal = auto()
    lower = auto()


class AcronymScheme(Enum):
    lower = auto()
    upper = auto()
    two_letter_limit = auto()


class Project:
    def __init__(
        self, words, requires_parameters=False, acronyms=None, acronym_scheme=None
    ):
        self._words = words
        self._requires_parameters = requires_parameters
        self._acronyms = [acronym.upper() for acronym in acronyms] if acronyms else []
        self._acronym_scheme = acronym_scheme or AcronymScheme.two_letter_limit

    @property
    def words(self):
        return self._words

    @property
    def requires_parameters(self):
        return self._requires_parameters

    @property
    def acronyms(self):
        return self._acronyms

    @property
    def acronym_scheme(self):
        return self._acronym_scheme

    @property
    def display_name(self):
        return self._as_display()

    def get_project_name_by_scheme(self, naming):
        """
        gets a project name for a specific naming scheme

        :param naming: the naming scheme
        :return: the project type formatted by the directory's naming scheme
        """
        try:
            return {
                NamingScheme.hyphen: self._as_hyphen(),
                NamingScheme.underscore: self._as_underscore(),
                NamingScheme.camel: self._as_camel(),
                NamingScheme.pascal: self._as_pascal(),
                NamingScheme.lower: self._as_lower(),
            }[naming]
        except KeyError as e:
            raise KeyError(f"Unknown naming scheme '{naming}'") from e

    def _as_hyphen(self):
        return "-".join(
            [self._try_as_acronym(word, NamingScheme.hyphen) for word in self._words]
        )

    def _as_underscore(self):
        return "_".join(
            [
                self._try_as_acronym(word, NamingScheme.underscore)
                for word in self._words
            ]
        )

    def _as_camel(self):
        return self._words[0].lower() + "".join(
            [
                self._try_as_acronym(word.title(), NamingScheme.camel)
                for word in self.words[1:]
            ]
        )

    def _as_pascal(self):
        return "".join(
            [
                self._try_as_acronym(word.title(), NamingScheme.pascal)
                for word in self._words
            ]
        )

    def _as_lower(self):
        return "".join([word.lower() for word in self._words])

    def _as_display(self):
        return " ".join(
            [
                self._try_as_acronym(word.title(), NamingScheme.underscore)
                for word in self._words
            ]
        )

    def _is_acronym(self, word):
        return word.upper() in self._acronyms

    def _try_as_acronym(self, word, naming_scheme):
        if self._is_acronym(word):
            if self._acronym_scheme == AcronymScheme.upper:
                return word.upper()
            elif self._acronym_scheme == AcronymScheme.lower:
                return word.lower()
            elif self._acronym_scheme == AcronymScheme.two_letter_limit:
                if len(word) <= 2 and naming_scheme in [
                    NamingScheme.camel,
                    NamingScheme.pascal,
                ]:
                    return word.upper()
        return word

    def __eq__(self, other):
        return (
            self._words == other.words
            and self._requires_parameters == other.requires_parameters
            and self._acronyms == other.acronyms
            and self._acronym_scheme == other.acronym_scheme
        )
